package grapheanim;

import java.awt.*;
import java.util.*;

/**
 * Description : Algorithme permettant de trouver le flot maximum d'un graphe.
 * @author <A href=http://www.alapetite.fr.st target=_blank>Alexandre Alapetite</A> & <A href=http://www.andujar.fr.st target=_blank>Brice Andujar</A>
 * @version 1.0
 */
public class FordFulkerson extends Methode_Generale
{
  /** graphe d'cart calcul  chaque tape */
  private GrapheData graphe_Ecart;

  public FordFulkerson(DynamicGraphData dgd)
  {
    super(dgd);
  }

  /** Cherche un chemin dans un graphe d'cart (rcursif).
   *  @param grapheDEcart Contient un graphe d'cart sans aller-retour entre 2 sommets.
   *  @param chemin Liste les sommets par lesquels passe le chemin,  l'envers: de l'arrive au dpart.
   *  @param verifCycle Pile permettant d'viter les cycles. Passer un stack vide ou un stack avec les sommets  pas utiliser.
   *  @param amelioration Amlioration maximale sur le chemin trouv. A initialiser  la valeur maximale tolrable.
   *  @param courant <CODE>Sommet</CODE> de dpart.
   *  @param destination <CODE>Sommet</CODE> d'arrive.
   *  @return vrai s'il existe un chemin de <CODE>courant</CODE>  <CODE>destination</CODE>, faux sinon.
   *  @see #grapheEcart
   */
  public boolean chercheCheminEcart(GrapheData grapheDEcart, Vector chemin, Stack verifCycle,PInteger amelioration, Sommet courant, Sommet destination)
  {
    int cpt=0;
    boolean trouve = false;
    if (verifCycle.contains(courant)) trouve=false; //on est en prsence d'un cycle, on abandonne ce chemin
    else
    {
      verifCycle.push(courant); //on ajoute le sommet courant  la liste des sommets utiliss pour le chemin test
      if ((courant==null) || (destination==null)) trouve=false;
      else if (courant.getNom().equals(destination.getNom()))
      { //on est arriv  destination
        trouve = true;
        //System.out.println("trouve"+amelioration.intValue());
        chemin.add(courant); //chemin crit en marche arrire
      }
      else while (!trouve && (cpt<courant.nbrSuccesseur()))
      {
        Successeur suivant = courant.getSuccesseurAt(cpt);
        if (!suivant.getNom().equals(courant.getNom())) //pas de retour sur le parent
        {
          trouve = chercheCheminEcart(grapheDEcart,chemin,verifCycle,amelioration,grapheDEcart.getSommet(suivant.getNom()),destination); //sinon on poursuit la recherche
          if (trouve)
          {
            if (suivant.getPoid()<amelioration.intValue())
              amelioration.setIntValue(suivant.getPoid()); //on garde l'amlioration minimale
            chemin.add(courant); //chemin crit en marche arrire
          }
        }
        cpt++;
      }
      if (!trouve) verifCycle.pop(); //le sommet actuel ne permet pas de trouver une chane amliorante, on le supprime de la liste des sommets utiliss
    }
    return trouve;
  }

  /** Range un <CODE>Vector</CODE> en sens inverse.
   *  @param verlan un <CODE>Vector</CODE>  l'envers.
   *  @return un <CODE>Vector</CODE>  l'endroit.
   */
  public Vector inverseVector(Vector verlan)
  {
    Vector lanver = new Vector(verlan.size());
    for (int cpt=verlan.size()-1; cpt>=0; cpt--) lanver.add(verlan.elementAt(cpt)); //remet chemin  l'endroit
    return lanver;
  }

  /** Construit un graphe d'cart  partir d'un graphe donn sans aller-retour entre 2 sommets.
   *  @param grapheData un graphe sans circuit.
   *  @param DD2 le DynamicData en cours (fait 0 addStep, 1 setAlgo)
   *  @return le graphe d'cart (bas sur les flux et les capacits) correspondant.
   *  @see DynamicGraphData#addStep
   */
  public GrapheData grapheEcart(GrapheData grapheData, DynamicData DD2)
  {
    GrapheData grapheDEcart = new GrapheData(grapheData);

    //Vector lesSommets = grapheDEcart.getAllSommets();
    for (int cpt=0; cpt<grapheDEcart.nbrSommet(); cpt++)
    {
      Sommet monSommet = grapheDEcart.sommetAt(cpt);
      for (int cpt2=monSommet.nbrSuccesseur()-1; cpt2>=0; cpt2--)
      {
        Successeur successeur = monSommet.getSuccesseurAt(cpt2);
        Sommet sommet2 = grapheDEcart.getSommet(successeur.getNom());
        //arc retour
        if (successeur.getFlot()>0) sommet2.addSuccesseur(monSommet.getNom(),successeur.getFlot());
        //arc de valeur: la diffrence entre la capacit et le flux
        int diff = successeur.getPoid()-successeur.getFlot();
        if (diff>0)
        {
          successeur.setPoid(diff);
          successeur.setFlot(0);
        }
        else monSommet.removeSuccesseur(successeur.getNom());
      }
    }

    String buffAffiche="Graphe d'cart:\n";
    for (int cpt=0; cpt<grapheDEcart.nbrSommet();cpt++)
    {
      Sommet mySommet = grapheDEcart.sommetAt(cpt);
      buffAffiche+=("Sommet "+mySommet.getNom()+":"+mySommet.getAllSuccesseurs().toString()+"\n");
    }
    DD2.setAlgo(buffAffiche+"\n");
    return grapheDEcart;
  }

  /** Amliore une chane : augmente (s'il sont dans le bon sens) ou diminue (sinon) les flux des arcs d'une chane.
   *  @param grapheData Graphe sur lequel doit se faire l'amlioration de flot.
   *  @param chemin liste des sommets (dans l'ordre) par lesquels passe la chane amliorante.
   *  @param amelioration valeur de l'amlioration.
   *  @param DD2 le DynamicData en cours (fait 2 addStep)
   *  @see #chercheCheminEcart
   */
  public void ameliorerChaine(GrapheData grapheData, Vector chemin, int amelioration, DynamicData DD2)
  {
    Sommet sommet1;
    Sommet sommet2;

    //Pour changer les artes du chemin en bleu
    for (int cpt=0; cpt<chemin.size()-1; cpt++)
    {
      sommet1 = grapheData.getSommet( ((Sommet)chemin.elementAt(cpt)).getNom() );
      sommet2 = grapheData.getSommet( ((Sommet)chemin.elementAt(cpt+1)).getNom() );
      Successeur success = sommet1.getSuccesseur(sommet2.getNom());
      if (success==null) //il n'y a pas ce successeur: donc arc dans le sens inverse
      {
        success = sommet2.getSuccesseur(sommet1.getNom());
        if (success!=null)
        {
          Sommet Old = sommet2.clonePoints();
          success.setCouleur(Color.blue);
          Sommet New = sommet2.clonePoints();
          DD2.addSommet(Old, New);
          //SauverCouleurArrete(DD2,Color.white,sommet2.getNom(),success.getNom());
        }
      }
      else //arc dans le bon sens
        {
          Sommet Old = sommet1.clonePoints();
          success.setCouleur(Color.blue);
          Sommet New = sommet1.clonePoints();
          DD2.addSommet(Old, New);
          //SauverCouleurArrete(DD2,Color.black,sommet1.getNom(),success.getNom());
        }
    }
    get_DGD().addStep(DD2);

    //amlioration et changement de couleur des arcs du chemin
    DD2 = new DynamicData();
    DD2.setCommentaire("Amlioration de "+amelioration+" de la chane "+chemin.toString()+"\n\n");
    for (int cpt=0; cpt<chemin.size()-1; cpt++)
    {
      sommet1 = grapheData.getSommet( ((Sommet)chemin.elementAt(cpt)).getNom() );
      sommet2 = grapheData.getSommet( ((Sommet)chemin.elementAt(cpt+1)).getNom() );
      Successeur success = sommet1.getSuccesseur(sommet2.getNom());
      if (success==null) //il n'y a pas ce successeur: donc arc dans le sens inverse
      {
        success = sommet2.getSuccesseur(sommet1.getNom());
        if (success!=null)
        {
          Sommet Old = sommet2.clonePoints();
          success.setFlot(success.getFlot()-amelioration);
          if (success.getFlot()<=0) success.setCouleur(Color.green);
          else success.setCouleur(new Color(0xD5FFD5));
          Sommet New = sommet2.clonePoints();
          DD2.addSommet(Old, New);
          //SauverCouleurArrete(DD2,Color.green,sommet2.getNom(),success.getNom());
        }
      }
      else //arc dans le bon sens
      {
        Sommet Old = sommet1.clonePoints();
        success.setFlot(success.getFlot()+amelioration);
        if (success.getFlot()>=success.getPoid()) success.setCouleur(Color.red);
        else success.setCouleur(Color.white);
        Sommet New = sommet1.clonePoints();
        DD2.addSommet(Old, New);
        //SauverCouleurArrete(DD2,Color.white,sommet1.getNom(),success.getNom());
      }
    }
    get_DGD().addStep(DD2);
  }

  /** Prparation,  appeler avant {@link #Explorer Explorer}. */
  public void Initialiser()
  {
    /*GrapheData GD1 = DGD.getInitGraph();
    for (int i = 0; i < GD1.nbrSommet(); i++)
      GD1.sommetAt(i).setPoid(100000);*/

    DynamicData DD =new DynamicData();
    DD.setCommentaire("Dbut de l'algorithme de Flots\n\n");
    get_DGD().addStep(DD);

    NOM(DGD.getOrigine()).set_Parent("NIL");    //on met le parent  'NIL'
  }

  /**
   * Droulement de tout l'algorithme de Flot.
   * Calcul du graphe d'cart, recherche une chane amliorante, amliore le chemin sur le graphe, re-graphe d'cart...
   */
  public void Explorer()
  {
    boolean encore=true; //indique s'il y a encore un chemin amliorant dans le graphe d'cart
    int augmenteTotal=0; //augmentation possible trouve dans le graphe d'cart sur un chemin
    GrapheData initGraphe = new GrapheData(DGD.getInitGraph());
    Sommet OldOld=NOM(DGD.getOrigine()).clonePoints();
    DynamicData DD;
    while (encore) //2 addStep par boucle
    {
      DD = new DynamicData();
      graphe_Ecart = grapheEcart(initGraphe,DD); //calcule le graphe d'cart
      Vector chemin = new Vector();
      PInteger amelioration = new PInteger(Integer.MAX_VALUE);
      //et cherche un chemin amliorant dans le graphe d'cart
      encore = chercheCheminEcart(graphe_Ecart,chemin,new Stack(),amelioration,graphe_Ecart.getSommet(DGD.getOrigine()),graphe_Ecart.getSommet(DGD.getArrivee()));
      if (encore)
      {
        chemin = inverseVector(chemin); //remet chemin  l'endroit
        //amliore la chane trouve grce au graphe d'cart
        ameliorerChaine(initGraphe,chemin,amelioration.intValue(),DD);
        //le addStep se fait dans ameliorerChaine
        augmenteTotal+=amelioration.intValue(); //incrmente l'amlioration totale apporte  ce graphe
      }
      else
      {
        //calcul de la valeur de flot obtenue
        int leFlot=0;
        Sommet lorigine = initGraphe.getSommet(DGD.getOrigine());
        for (int cpt=0; cpt<lorigine.nbrSuccesseur(); cpt++)
          leFlot+=lorigine.getSuccesseurAt(cpt).getFlot();

        DD.setCommentaire("Plus de chemin amliorant.\n"+
                          "Le flot max est de "+leFlot+". Amlioration de "+augmenteTotal+".\n"+
                          "Les arcs blancs ont t amliors, les rouges saturs,\n"+
                          "les vert-pomme diminus et verts fluo diminus  0");
        get_DGD().addStep(DD);
      }
    }
  }

}
